﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace RarMount2
{
    class RarCollection
    {
        private RarFile[] RarFiles;
        private string strRarFileName;   // <- name of stored file
        private long lngRarFileFullSize; // <- full file size of stored file
        private bool boolFilesLoadedOk = false; // <- indicates the status

        public RarCollection(string strMainFileName)
        {
            boolFilesLoadedOk = findArchives(strMainFileName); // find spanned files and load info
        }

        // N.B! Offset is reading offset, not returning offset.
        public int readData(byte[] byteRarReadBuffer, ulong lngReadOffset, int intReadLength)
        {
            int intByteFilled = 0;
            int intReadMax = 0;
            long lngCurrPosition;
            foreach (RarFile rfCurrRar in RarFiles)
            {
                if (lngReadOffset >= rfCurrRar.getRarFileOffset() && lngReadOffset < rfCurrRar.getRarFileOffset() + (ulong)rfCurrRar.getRarFileSize())
                {
                    // Requested data starts within this file
                    lngCurrPosition = (long)lngReadOffset - (long)rfCurrRar.getRarFileOffset() + rfCurrRar.getRarFileStart();
                    if (lngCurrPosition - rfCurrRar.getRarFileStart() + (long)intReadLength > rfCurrRar.getRarFileSize())
                        intReadMax = (int)(rfCurrRar.getRarFileSize() - lngCurrPosition + rfCurrRar.getRarFileStart());
                    else
                        intReadMax = intReadLength;
                    intByteFilled = rfCurrRar.readData(lngCurrPosition, byteRarReadBuffer, intByteFilled, intReadMax);
                }
                else if (intByteFilled > 0 && intByteFilled < intReadLength)
                {
                    // Last file didn't contain all requested data
                    lngCurrPosition = rfCurrRar.getRarFileStart();
                    if (intReadLength - intByteFilled > rfCurrRar.getRarFileSize())
                        intReadMax = (int)rfCurrRar.getRarFileSize();
                    else
                        intReadMax = intReadLength - intByteFilled;
                    intByteFilled = intByteFilled + rfCurrRar.readData(lngCurrPosition, byteRarReadBuffer, intByteFilled, intReadMax);
                }
                else if (intByteFilled == intReadLength)
                {
                    // All data is here :)
                    return intByteFilled;
                }
            }
            // Didn't get all data. Return it anyway.
            return intByteFilled;
        }

        public int getLoadedFilesCount()
        {
            return RarFiles.Length;
        }

        public bool filesLoadedOk()
        {
            return boolFilesLoadedOk;
        }

        public string getStoredFileName()
        {
            return strRarFileName;
        }

        public long getStoredFileSize()
        {
            return lngRarFileFullSize;
        }

        private bool findArchives(string strStartingFile)
        {

            FileInfo fiMain = new FileInfo(strStartingFile);
            string strFileSearchPattern = "";
            if ((fiMain.Name.Length > 12 && fiMain.Name.ToLower().Substring(fiMain.Name.Length - 12, 12) == ".part001.rar") ||
                (fiMain.Name.Length > 11 && fiMain.Name.ToLower().Substring(fiMain.Name.Length - 11, 11) == ".part01.rar")) // Fixed bug 2008-05-24 (found with Kyle.XY.S02E17.REAL.REPACK.DSR.XviD-0TV)
            {
                frmMain.debugPrint("DEBUG - File ending is '.part001.rar' or '.part01.rar', will be looking for '.rar' files.");
                strFileSearchPattern = fiMain.Name.Substring(0, fiMain.Name.Length - 7) + "*.rar";
            }
            else if (fiMain.Name.ToLower().Substring(fiMain.Name.Length - 4, 4) == ".rar")
            {
                frmMain.debugPrint("DEBUG - File ending is '.rar', will be looking for '.r00', '.r01', etc.");
                strFileSearchPattern = fiMain.Name.Substring(0, fiMain.Name.Length - 4) + ".r*";
            }
            else
            {
                // FIXME: Handle this situation!
                frmMain.debugPrint("DEBUG - Unknown file name ending (should be '.part001.rar', '.part01.rar' or '.rar'): " + strStartingFile);
            }


            FileInfo[] fiFiles = fiMain.Directory.GetFiles(strFileSearchPattern);
            RarFiles = new RarFile[fiFiles.Length];
            RarFiles[0] = new RarFile(fiMain.FullName, 0);
            if (!RarFiles[0].loadFileInfo())
            {
                frmMain.debugPrint("DEBUG - Failed to load file info from archive: " + fiMain.FullName);
                return false; // <- error occured in loadFileInfo()
            }

            strRarFileName = RarFiles[0].getRarFileName();
            lngRarFileFullSize = RarFiles[0].getRarFileFullSize();

            int intRarNo = 1;
            foreach (FileInfo fiPart in fiFiles)
            {
                if (fiPart.Name != fiMain.Name)
                {
                    RarFiles[intRarNo] = new RarFile(fiPart.FullName, (long)RarFiles[intRarNo - 1].getRarFileOffset() + RarFiles[intRarNo - 1].getRarFileSize());
                    if (!RarFiles[intRarNo].loadFileInfo())
                    {
                        frmMain.debugPrint("DEBUG - Failed to load file info from archive: " + fiPart.FullName);
                        return false; // <- error occured in loadFileInfo()
                    }
                    intRarNo++;
                }
            }

            return true;
        }
    }
}
