﻿using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace RarMount2
{
    // Simulates a 31 GiB NTFS harddrive, based on reverse engineering
    /* INFO
     * A 31GiB NTFS partition with one file, 15 GiB UNPACKED.RAW, looks something like this:
     * Byte from    Byte to     Description
     * 0            1476095     NTFS header, device info, etc.
     * 22446080     22450175    INDX record, dictonary info?
     * 3154083840   3154710527  RSTR record and some (unused?) file info
     * 3154710528   3221192703  Probably allocation info, in this case every byte is hex: FF
     * 3221192704   3221270527  File info again, incl file name and file size
     * 16642965504  16643391487 File info again, incl file name and file size
     * 16643391488  32749518847 File UNPACKED.RAW 15 GiB data
     * 33285996032  33285996543 NTFS footer
     * Everywhere else          Every byte is hex: 00
     */
    class SimpleNTFS
    {
        private string strBinaryHelpFile = "";
        private byte[] byte15gbFileType;
        private long lng15gbFileSize = 0;
        RarCollection rcMain;
        private FileStream fsHelpFileReader;

        /* FIXME: Future support for SRT/SUB
        public SimpleNTFS(string strHelpFile, string strUncompressedFile, string strSubtitleFile)
        {
        }*/

        public SimpleNTFS(string strHelpFile, RarCollection rarUncompressedFile)
        {
            strBinaryHelpFile = strHelpFile;
            rcMain = rarUncompressedFile;
            if (rcMain.filesLoadedOk())
            {
                try
                {
                    fsHelpFileReader = new FileStream(strBinaryHelpFile, FileMode.Open, FileAccess.Read);
                    string str15gbFileName = rcMain.getStoredFileName();
                    string str15gbFileType = str15gbFileName.Substring(str15gbFileName.Length - 3, 3).ToUpper();
                    byte15gbFileType = Encoding.ASCII.GetBytes(str15gbFileType);
                    lng15gbFileSize = rcMain.getStoredFileSize();
                }
                catch
                {
                }
            }
        }

        public bool loadedOk()
        {
            // We should do some more checks here ...
            try
            {
                if (lng15gbFileSize > 0 && fsHelpFileReader.Length > 0)
                    return true;
                else
                    return false;
            }
            catch
            {
                return false;
            }
        }

        public int readData(byte[] byteNtfsReadBuffer, ulong lngNtfsReadOffset, int intReadLength)
        {
            // FIXME:
            //  Precache byte arrays with file size and name (no need to regenerate every time accessed)
            //  Comments!

            int intReadMax;
            int intBytesLeftToRead;
            int intBytesReadTotal = 0;
            ulong lngReadOffset;
            while (intReadLength > intBytesReadTotal)
            {
                intBytesLeftToRead = intReadLength - intBytesReadTotal;
                lngReadOffset = lngNtfsReadOffset + (ulong)intBytesReadTotal;
                if (lngReadOffset >= 16643391488 && lngReadOffset < 16643391488 + (ulong)lng15gbFileSize)
                {
                    if (lngReadOffset + (ulong)intBytesLeftToRead > 16643391488 + (ulong)lng15gbFileSize)
                        intReadMax = (int)((16643391488 + (ulong)lng15gbFileSize) - lngReadOffset);
                    else
                        intReadMax = intBytesLeftToRead;
                    ulong lngStartPosition = lngReadOffset - 16643391488;
                    int intBytesRead = 0;
                    intBytesRead = rcMain.readData(byteNtfsReadBuffer, lngStartPosition, intReadMax);
                    // Fixed bug 2008-05-24, avoid never ending loop
                    if (intBytesRead == 0)
                        break;
                    // /Fixed bug

                    intBytesReadTotal += intBytesRead;
                }
                else if (lngReadOffset >= 0 && lngReadOffset < 1476096)
                {
                    intBytesReadTotal = ntfsHelpFileRead(0, 0, 1476096, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                }
                else if (lngReadOffset >= 22446080 && lngReadOffset < 22450176)
                {
                    intBytesReadTotal = ntfsHelpFileRead(1476096, 22446080, 22450176, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                }
                else if (lngReadOffset >= 3154083840 && lngReadOffset < 3154710528)
                {
                    intBytesReadTotal = ntfsHelpFileRead(1480192, 3154083840, 3154710528, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                }
                else if (lngReadOffset >= 3221192704 && lngReadOffset < 3221269764) //(lngReadOffset >= 3221192704 && lngReadOffset < 3221270528)
                {
                    intBytesReadTotal = ntfsHelpFileRead(2106880, 3221192704, 3221269764, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                }
                else if (lngReadOffset >= 3221269764 && lngReadOffset < 3221269840)
                {
                    // File name and size (x2)
                    byte[] byteFileSizeData = new byte[76];
                    BitConverter.GetBytes(lng15gbFileSize).CopyTo(byteFileSizeData, 60);
                    BitConverter.GetBytes(lng15gbFileSize).CopyTo(byteFileSizeData, 68);
                    byteFileSizeData[0] = byte15gbFileType[0]; // A
                    byteFileSizeData[2] = byte15gbFileType[1]; // V
                    byteFileSizeData[4] = byte15gbFileType[2]; // I
                    byteFileSizeData[12] = 0x80;
                    byteFileSizeData[16] = 0x48;
                    byteFileSizeData[20] = 0x01;
                    byteFileSizeData[26] = 0x01;
                    byteFileSizeData[36] = 0xFF;
                    byteFileSizeData[37] = 0x7F;
                    byteFileSizeData[38] = 0x07;
                    byteFileSizeData[44] = 0x40;
                    byteFileSizeData[55] = 0xC0;
                    byteFileSizeData[56] = 0x03;


                    if (lngReadOffset + (ulong)intBytesLeftToRead > 3221269840)
                        intReadMax = (int)(3221269840 - lngReadOffset);
                    else
                        intReadMax = intBytesLeftToRead;

                    int intStartPosition = (int)(lngReadOffset - 3221269764);
                    for (int i = 0; i < intReadMax; i++)
                    {
                        byteNtfsReadBuffer[i + intBytesReadTotal] = byteFileSizeData[i + intStartPosition];
                    }
                    intBytesReadTotal += intReadMax;
                }
                else if (lngReadOffset >= 3221269840 && lngReadOffset < 3221270528)
                {
                    intBytesReadTotal = ntfsHelpFileRead(2184016, 3221269840, 3221270528, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                }
                else if (lngReadOffset >= 16642965504 && lngReadOffset < 16643065136) //(lngReadOffset >= 16642965504 && lngReadOffset < 16643391488)
                {
                    intBytesReadTotal = ntfsHelpFileRead(2184704, 16642965504, 16643065136, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                }
                else if (lngReadOffset >= 16643065136 && lngReadOffset < 16643065177)
                {
                    // File name and size
                    byte[] byteFileSizeData = new byte[41];
                    BitConverter.GetBytes(lng15gbFileSize).CopyTo(byteFileSizeData, 0);
                    byteFileSizeData[8] = 0x20;
                    byteFileSizeData[16] = 0x0C;
                    byteFileSizeData[17] = 0x03;
                    byteFileSizeData[18] = 0x55;
                    byteFileSizeData[20] = 0x4E;
                    byteFileSizeData[22] = 0x50;
                    byteFileSizeData[24] = 0x41;
                    byteFileSizeData[26] = 0x43;
                    byteFileSizeData[28] = 0x4B;
                    byteFileSizeData[30] = 0x45;
                    byteFileSizeData[32] = 0x44;
                    byteFileSizeData[34] = 0x2E;
                    byteFileSizeData[36] = byte15gbFileType[0]; // A
                    byteFileSizeData[38] = byte15gbFileType[1]; // V
                    byteFileSizeData[40] = byte15gbFileType[2]; // I


                    if (lngReadOffset + (ulong)intBytesLeftToRead > 16643065177)
                        intReadMax = (int)(16643065177 - lngReadOffset);
                    else
                        intReadMax = intBytesLeftToRead;

                    int intStartPosition = (int)(lngReadOffset - 16643065136);
                    for (int i = 0; i < intReadMax; i++)
                    {
                        byteNtfsReadBuffer[i + intBytesReadTotal] = byteFileSizeData[i + intStartPosition];
                    }
                    intBytesReadTotal += intReadMax;
                }
                else if (lngReadOffset >= 16643065177 && lngReadOffset < 16643391488)
                {
                    intBytesReadTotal = ntfsHelpFileRead(2284377, 16643065177, 16643391488, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                }
                else if (lngReadOffset >= 33285996032 && lngReadOffset < 33285996544)
                {
                    intBytesReadTotal = ntfsHelpFileRead(2610688, 33285996032, 33285996544, byteNtfsReadBuffer, lngReadOffset, intBytesReadTotal, intBytesLeftToRead);
                    // Last part of file, no more data to read
                    break;
                }
                else if (lngReadOffset >= 3154710528 && lngReadOffset < 3221192704)
                {
                    // Return hex(FF) filled array
                    if (lngReadOffset + (ulong)intBytesLeftToRead > 3221192704)
                        intReadMax = (int)(3221192704 - lngReadOffset);
                    else
                        intReadMax = intBytesLeftToRead;
                    for (int i = intBytesReadTotal; i < intBytesReadTotal + intReadMax; i++)
                    {
                        byteNtfsReadBuffer[i] = 255;
                    }
                    intBytesReadTotal += intReadMax;
                }
                else
                {
                    // Return zero filled array
                    intReadMax = 1; // Slow but safest
                    for (int i = intBytesReadTotal; i < intBytesReadTotal + intReadMax; i++)
                    {
                        byteNtfsReadBuffer[i] = 0;
                    }
                    intBytesReadTotal += intReadMax;
                }
            }
            return intBytesReadTotal;
        }

        private int ntfsHelpFileRead(int intHelpFileOffset, ulong lngLowerOffset, ulong lngUpperOffset, byte[] byteNtfsReadBuffer, ulong lngReadOffset, int intBytesReadTotal, int intBytesLeftToRead)
        {
            int intReadMax;
            if (lngReadOffset + (ulong)intBytesLeftToRead > lngUpperOffset)
                intReadMax = (int)(lngUpperOffset - lngReadOffset);
            else
                intReadMax = intBytesLeftToRead;
            fsHelpFileReader.Position = (long)(lngReadOffset - lngLowerOffset) + (long)intHelpFileOffset;
            int intBytesRead = fsHelpFileReader.Read(byteNtfsReadBuffer, intBytesReadTotal, intReadMax);
            return intBytesReadTotal + intBytesRead;
        }
    }
}
